from abc import abstractmethod
from typing import Any

from returns.future import FutureResult


class EventHandlingTracker:
    @abstractmethod
    def check_if_notif_handled(self, a_message_id: str) -> FutureResult[bool, Any]:
        pass

    @abstractmethod
    def mark_notif_as_handled(self, a_message_id: str) -> FutureResult:
        pass



from typing import TypeVar, overload

NumberType = TypeVar("NumberType")

# Not pure
def increase(a_number: NumberType, amount: NumberType) -> NumberType:
    return a_number + amount


def negate(a_bool: bool) -> bool:
    return not a_bool


@overload
def is_in_range(a_value: int, a_minimum: int, a_maximum: int) -> bool:
    ...


@overload
def is_in_range(a_value: float, a_minimum: float, a_maximum: float) -> bool:
    ...


def is_in_range(a_value, a_minimum, a_maximum):
    return a_value >= a_minimum and a_value <= a_maximum



from typing import Optional

from .DomainEventPublisher import DomainEventPublisher

# __all__ = ["DomainService"]


class DomainService:
    """
    Domain services carry domain knowledge that doesnt naturally fit entities and value objects.
    """

    _domain_event_publisher: DomainEventPublisher | None

    def __init__(
        self,
        domain_event_publisher: Optional[DomainEventPublisher] = None,
    ) -> None:
        self._domain_event_publisher = domain_event_publisher
        super().__init__()

    def domain_event_publisher(self) -> DomainEventPublisher:
        return self._domain_event_publisher or DomainEventPublisher.instance()



import json
from typing import Any, List, TypedDict, TypeVar

from dino_seedwork_be.utils.dict import extract

from ..utils import Filter

FilterType = TypeVar("FilterType", bound=TypedDict)

__all__ = ["Filters"]


class Filters:
    def __init__(self, keys: List[str]) -> None:
        self.keys = keys

    def __call__(self, filters: Any = {}):
        match filters:
            case str():
                try:
                    filters = json.loads(filters)
                except Exception as error:
                    filters = {}
        plain_filter = extract(filters, self.keys)
        return Filter(plain_filter).parsed_filter



from .AbstractJsonMediaReader import AbstractJSONMediaReader, JSONReader
from .RepresentationReader import RepresentationReader

__all__ = ["JSONReader", "AbstractJSONMediaReader", "RepresentationReader"]



from dataclasses import field
from typing import Generic, TypeVar

from returns.maybe import Maybe, Nothing
from returns.result import safe

from dino_seedwork_be.domain.value_object.AbstractIdentity import \
    AbstractIdentity

from .DomainAssertionConcern import DomainAssertionConcern

# __all__ = ["IdentifiedDomainObject"]

IdentityType = TypeVar("IdentityType", bound=AbstractIdentity)


class IdentifiedDomainObject(Generic[IdentityType], DomainAssertionConcern):
    _id: Maybe[IdentityType]

    def __eq__(self, obj):
        return self._id == obj.id

    def __init__(self, id: Maybe[IdentityType] = Nothing):
        self.set_id(id)
        super().__init__()

    def identity(self) -> Maybe[IdentityType]:
        return self._id

    def id_as_string(self) -> Maybe[str]:
        return self._id.map(lambda id: id.get_raw_str())

    @safe
    def set_id(self, id: Maybe[IdentityType]):
        self._id = id



from typing import Any, List, Optional

from pika.exchange_type import ExchangeType
from returns.curry import partial
from returns.functions import tap
from returns.future import FutureFailure, FutureResult, FutureSuccess
from returns.iterables import Fold
from returns.maybe import Maybe
from returns.pipeline import flow, managed, pipe
from returns.pointfree import bind, map_
from returns.result import Result, Success

from dino_seedwork_be.adapters.messaging.notification import (
    Notification, NotificationPublisher, NotificationSerializer,
    PublishedNotificationTrackerStore)
from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import \
    SuperDBSessionUser
from dino_seedwork_be.domain.event.EventSerializer import EventSerializer
from dino_seedwork_be.domain.event.EventStore import EventStore
from dino_seedwork_be.domain.event.StoredEvent import StoredEvent
from dino_seedwork_be.exceptions import MainException
from dino_seedwork_be.utils import (feed_args, feed_kwargs,
                                    print_result_with_text)
from dino_seedwork_be.utils.functional import return_v

from . import (RabbitMQConnectionSettings, RabbitMQExchange,
               RabbitMQMessageParameters, RabbitMQMessageProducer)

# __all__ = ["RabbitMQPublisher"]


class RabbitMQPublisher(NotificationPublisher, SuperDBSessionUser):
    _event_store: EventStore
    _exchange_name: str
    _exchange_type: ExchangeType
    _published_notif_tracker_store: PublishedNotificationTrackerStore
    _message_producer_ins: Optional[RabbitMQMessageProducer] = None
    _event_serializer: EventSerializer
    _connection_settings: RabbitMQConnectionSettings

    def __init__(
        self,
        event_store: EventStore,
        exchange_name: str,
        exchange_type: ExchangeType,
        published_notif_tracker_store: PublishedNotificationTrackerStore,
        event_serializer: EventSerializer,
        connection_settings: RabbitMQConnectionSettings,
    ) -> None:
        # session = session_factory()
        self.set_exchange_name(exchange_name)
        self.set_event_store(event_store)
        # self._event_store.set_session(session)
        self._connection_settings = connection_settings

        def init_message_producer():
            self._exchange_type = exchange_type
            self._message_producer()

        init_message_producer()
        self.set_published_notif_tracker_store(published_notif_tracker_store)
        # self._published_notif_tracker_store.set_session(session)
        self.set_session_users(
            [self.event_store(), self.published_notif_tracker_store()]
        )
        self._event_serializer = event_serializer
        super().__init__()

    def event_serializer(self):
        return self._event_serializer

    def publish_notifications(self) -> FutureResult[Maybe[int], Any]:
        """
        publish unpublished event in event store
        :return: return the last event int id that published already
        """

        def notifications_from(
            events: List[StoredEvent],
        ) -> Result[List[Notification], Any]:
            return flow(
                events,
                partial(
                    map,
                    lambda stored_event: flow(
                        None,
                        lambda _: self.event_serializer().deserialize(
                            stored_event.body()
                        ),
                        map_(
                            lambda domain_event: list(
                                [
                                    stored_event.id().value_or(None),
                                    domain_event,
                                ]
                            )
                        ),
                        bind(feed_args(Notification.factory)),
                    ),
                ),
                list,
                lambda results: Fold.collect(results, Success(())),
            )

        def publish_future(
            msg_producer: RabbitMQMessageProducer,
        ) -> FutureResult[Maybe[int], Any]:
            match msg_producer.is_ready_for_publish():
                case False:
                    return FutureFailure(
                        MainException("RabbitMQMessageProducer not ready yet")
                    )
                case True:
                    return (
                        self.published_notif_tracker_store()
                        .published_notification_tracker()
                        .bind(
                            lambda tracker: flow(
                                tracker.most_recent_published_notification_id().value_or(
                                    0
                                ),
                                self.event_store().all_stored_events_since,
                                bind(
                                    pipe(notifications_from, FutureResult.from_result)
                                ),
                                map_(
                                    print_result_with_text("notification from results")
                                ),
                                bind(
                                    pipe(
                                        partial(
                                            map,
                                            lambda notif: flow(
                                                notif,
                                                partial(self._publish, msg_producer),
                                                map_(return_v(notif)),
                                            ),
                                        ),
                                        lambda publish_results: Fold.collect(
                                            publish_results, FutureSuccess(())
                                        ),
                                    )
                                ),
                                bind(
                                    partial(
                                        self.published_notif_tracker_store().track_most_recent_published_notification,
                                        tracker,
                                    )
                                ),
                            )
                        )
                    )

        def close_producer(producer: RabbitMQMessageProducer, _):
            producer.close()
            return FutureSuccess(None)

        return managed(publish_future, close_producer)(
            FutureResult.from_result(self._message_producer())
        )

    def _notification_routing_key(self, notification: Notification):
        return notification.type_name()

    def _publish(
        self,
        a_message_producer: RabbitMQMessageProducer,
        a_notification: Notification,
    ) -> FutureResult[None, Exception]:
        return flow(
            {
                "a_type": a_notification.type_name(),
                "a_message_id": str(a_notification.id()),
                "a_timestamp": int(a_notification.occurred_on().timestamp()),
            },
            feed_kwargs(RabbitMQMessageParameters.durable_text_parameters),
            lambda text_parameters: flow(
                a_notification,
                NotificationSerializer.instance().serialize,
                map_(tap(print_result_with_text("notification serialized"))),
                lambda notif_string: FutureResult.from_result(
                    a_message_producer.send(
                        text_parameters,
                        notif_string,
                        self._notification_routing_key(a_notification),
                    )
                ),
            ),
        )

    def event_store(self) -> EventStore:
        return self._event_store

    def exchange_name(self) -> str:
        return self._exchange_name

    def exchange_type(self) -> ExchangeType:
        return self._exchange_type

    def published_notif_tracker_store(self) -> PublishedNotificationTrackerStore:
        return self._published_notif_tracker_store

    def set_event_store(self, event_store: EventStore):
        self._event_store = event_store

    def set_exchange_name(self, a_name: str):
        self._exchange_name = a_name

    def _set_message_producer(self, message_producer: RabbitMQMessageProducer):
        self._message_producer_ins = message_producer

    def _message_producer(self) -> Result[RabbitMQMessageProducer, Any]:
        match self._message_producer_ins:
            case None:
                return flow(
                    [
                        self._connection_settings,
                        self.exchange_name(),
                        self.exchange_type(),
                        True,
                    ],
                    feed_args(RabbitMQExchange.factory),
                    map_(RabbitMQMessageProducer.factory),
                    map_(tap(self._set_message_producer)),
                )
            case RabbitMQMessageProducer():
                return Success(self._message_producer_ins)
            case _:
                print("not message producer in case", self._message_producer_ins)

    def set_published_notif_tracker_store(
        self, a_published_notif_tracker_store: PublishedNotificationTrackerStore
    ):
        self._published_notif_tracker_store = a_published_notif_tracker_store

    def run(self):
        self._message_producer().unwrap().run()

    def is_ready(self) -> Result[bool, Any]:
        return (
            self._message_producer()
            .map(lambda producer: producer.is_ready_for_publish())
            .lash(lambda _: Success(False))
        )

    def get_last_published_notification_id(self) -> FutureResult[Maybe[int], Any]:
        return (
            self.published_notif_tracker_store()
            .published_notification_tracker()
            .map(lambda tracker: tracker.most_recent_published_notification_id())
        )



from .IDto import DTO
from .IRepository import PaginationResultDB, Repository
from .IRepositoryUUID import RepositoryUUID
from .Mapper import Mapper

__all__ = ["DTO", "PaginationResultDB", "Repository", "RepositoryUUID", "Mapper"]



from .AbstractProcess import IProcess, ProcessCompletionType
from .AbstractProcessApplicationService import \
    AbstractProcessApplicationService
from .ProcessId import ProcessId
from .ProcessTimeOut import ProcessTimedOut
from .TimeConstrainedProcessTracker import TimeConstrainedProcessTracker
from .TimeConstrainedProcessTrackerRepository import \
    TimeConstrainedProcessTrackerRepository
from .timeout_event_factory import factory_timeout_event

__all__ = [
    "ProcessCompletionType",
    "IProcess",
    "AbstractProcessApplicationService",
    "ProcessId",
    "ProcessTimedOut",
    "TimeConstrainedProcessTracker",
    "TimeConstrainedProcessTrackerRepository",
    "factory_timeout_event",
]



from .AbstractUnitOfWork import AbstractUnitOfWork
from .DBSessionUser import (DBSessionUser, SessionUserAlreadyHaveSession,
                            SuperDBSessionUser)

__all__ = [
    "AbstractUnitOfWork",
    "DBSessionUser",
    "SessionUserAlreadyHaveSession",
    "SuperDBSessionUser",
]



from typing import Callable, List

from dino_seedwork_be.adapters.persistance.sql.AbstractUnitOfWork import \
    AbstractUnitOfWork
from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import \
    DBSessionUser

UowFactory = Callable[[List[DBSessionUser]], AbstractUnitOfWork]

# __all__ = ["AbstractUOWApplicationService"]


class AbstractUOWApplicationService:

    _uow: UowFactory

    def uow(self, user_session: List[DBSessionUser]) -> AbstractUnitOfWork:
        return self._uow(user_session)

    def set_uow(self, uow: UowFactory):
        self._uow = uow



from typing import List

from returns._internal.pipeline.flow import flow
from returns.maybe import Maybe
from toolz.dicttoolz import assoc, get_in
from toolz.functoolz import reduce

from dino_seedwork_be.utils.functional import feed_kwargs


def keys(v: dict):
    return v.keys()


def values(v: dict):
    return v.values()


def dict_to_cls(v: dict, keys: List[str], Cls):
    return flow(
        reduce(lambda acc, key: assoc(acc, key, v.get(key, None)), keys, {}),
        feed_kwargs(Cls),
    )


def extract(v: dict, keys: List[str]) -> dict:
    return {k: v.get(k, None) for k in keys}


def path(paths: List[str], v: dict) -> Maybe:
    return Maybe.from_optional(get_in(paths, v, None))



import enum
from typing import List, Optional

from returns.maybe import Maybe


class NotImplementError(RuntimeError):
    def __init__(self, message: str):
        super().__init__(f"NotImplementError: {message}")


class ExceptionCode(enum.Enum):
    RAW_EXCEPTION = "RAW_EXCEPTION"


class MainException(Exception):
    _code: Optional[str] = ExceptionCode.RAW_EXCEPTION.value
    _message: Optional[str] = None
    _loc: Optional[List[str]] = None

    def __init__(
        self,
        code: Optional[str] = None,
        message: Optional[str] = None,
        loc: Optional[List[str]] = None,
    ):
        super().__init__(f"[{code}]: {message} at {loc}")
        self._code = code
        self._message = message
        self._loc = loc

    def message(self) -> Maybe[str]:
        return Maybe.from_optional(self._message)

    def set_loc(self, loc: List[str]):
        self._loc = loc

    def loc(self) -> Maybe[List[str]]:
        return Maybe.from_optional(self._loc)

    def code(self) -> Maybe[str]:
        return Maybe.from_optional(self._code)

    def where(self) -> Maybe[List[str]]:
        return Maybe.from_optional(self._loc)


class IllegalArgumentException(MainException):
    def __init__(
        self,
        message: Optional[str],
        loc: Optional[List[str]] = None,
        code: Optional[str] = None,
    ):
        super().__init__(
            message=f"IllegalArgumentException: {message} - loc: {loc}",
            loc=loc,
            code=code,
        )


class IllegalStateException(MainException):
    def __init__(
        self,
        message: Optional[str],
        loc: Optional[List[str]] = None,
        code: Optional[str] = None,
    ):
        super().__init__(
            message=f"IllegalStateException: {message} - loc: {loc}", loc=loc, code=code
        )


def except_locs(locs: List[str]):
    def decor(function):
        def wrapper(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except MainException as error:
                current_locs = error.loc().value_or([])
                if len(locs) <= len(current_locs):
                    is_begin_sub_array = all(
                        [elem == current_locs[idx] for idx, elem in enumerate(locs)]
                    )
                    if is_begin_sub_array:
                        raise error
                error.set_loc([*locs, *current_locs])
                raise error

        return wrapper

    return decor



from .DefaultImplementNotificationApplicationService import \
    DefaultImplementNotificationApplicationService
from .Notification import DomainEventSerializable, Notification
from .NotificationPublisher import NotificationPublisher
from .NotificationReader import NotificationReader
from .NotificationSerializer import NotificationSerializer
from .PublishedNotificationTracker import PublishedNotificationTracker
from .PublishedNotificationTrackerStore import \
    PublishedNotificationTrackerStore

__all__ = [
    "DefaultImplementNotificationApplicationService",
    "DomainEventSerializable",
    "Notification",
    "NotificationPublisher",
    "NotificationReader",
    "NotificationSerializer",
    "PublishedNotificationTracker",
    "PublishedNotificationTrackerStore",
]



from typing import Optional

from pika.credentials import PlainCredentials
from returns.maybe import Maybe, Nothing, Some

# __all__ = ["ConnectionSettings"]


class ConnectionSettings:
    _host_name: str
    _port: Optional[int]
    _username: Optional[str]
    _password: Optional[str]

    # My virtualHost, which is the name of the RabbitMQ virtual host
    _virtual_host: str

    def __init__(
        self,
        a_host_name: str,
        a_port: int,
        a_virtual_host: str,
        an_username: Optional[str],
        a_password: Optional[str],
    ) -> None:
        self._host_name = a_host_name
        self._password = a_password
        self._port = a_port
        self._username = an_username
        self._virtual_host = a_virtual_host

    def host_name(self) -> str:
        return self._host_name

    def password(self) -> Maybe[str]:
        return Maybe.from_optional(self._password)

    def port(self) -> Maybe[int]:
        return Maybe.from_optional(self._port)

    def username(self) -> Maybe[str]:

        return Maybe.from_optional(self._username)

    def virtual_host(self) -> str:
        return self._virtual_host

    @staticmethod
    def factory(a_host_name, a_port, a_virtual_host, an_username, a_password):
        return ConnectionSettings(
            a_host_name=a_host_name,
            a_port=a_port,
            a_virtual_host=a_virtual_host,
            an_username=an_username,
            a_password=a_password,
        )

    def has_user_credentials(self):
        match [self.password(), self.username()]:
            case [Some(), Some()]:
                return True
            case _:
                return False

    def get_credential(self) -> Maybe[PlainCredentials]:
        match [self.password(), self.username()]:
            case [Some(), Some()]:
                return Some(
                    PlainCredentials(
                        username=self.username().unwrap(),
                        password=self.password().unwrap(),
                    )
                )
            case _:
                return Nothing



from datetime import datetime

from returns.maybe import Maybe, Nothing, Some

from dino_seedwork_be.domain.value_object.AbstractValueObject import \
    ValueObject

from .IdentifiedDomainObject import IdentifiedDomainObject, IdentityType

# __all__ = ["IdentifiedValueObject"]


class IdentifiedValueObject(ValueObject, IdentifiedDomainObject[IdentityType]):
    """
    Base class for identified objects
    """

    def __init__(self, an_id: IdentityType, created_at: Maybe[datetime] = Nothing):
        self.set_id(Some(an_id))
        super().__init__(created_at=created_at)



import asyncio
import traceback
from collections.abc import Callable, Iterable
from typing import Any, Coroutine, List, Optional, ParamSpec, TypeVar

from returns.converters import maybe_to_result
from returns.curry import curry, partial
from returns.future import (Future, FutureFailure, FutureResult, FutureResultE,
                            future_safe)
from returns.io import IOFailure, IOResult, IOSuccess
from returns.iterables import Fold
from returns.maybe import Maybe
from returns.pipeline import flow, is_successful
from returns.pointfree import bind, lash
from returns.primitives.exceptions import UnwrapFailedError
from returns.result import Failure, Result, Success
from returns.unsafe import unsafe_perform_io

from dino_seedwork_be.exceptions import MainException

T = TypeVar("T")

ExceptionType = TypeVar("ExceptionType", bound=Exception)


def for_each(function: Callable[[T, int], Any], iterable: Iterable[T]):
    for idx, element in enumerate(iterable):
        function(element, idx)


def map_to_list(fn: Callable[[T], Any], items: List[T]) -> List[Any]:
    return flow(items, partial(map, fn), list)


def feed_identity(obj):
    return lambda _: obj


def get_class_name(obj):
    return obj.__class__.__name__


def set_private_attr(obj, field, value):
    setattr(obj, f"_{get_class_name(obj)}__{field}", value)


def set_public_attr(obj, field, value):
    setattr(obj, field, value)


def set_protected_attr(obj, field, value):
    setattr(obj, field, value)


InnerValueType = TypeVar("InnerValueType")


def maybe_to_future(v: InnerValueType) -> Future[Result[InnerValueType, Any]]:
    return flow(
        v,
        Maybe.from_optional,
        maybe_to_result,
        FutureResult.from_result,
        Future.from_future_result,
    )


def result_to_future(
    v: Result[InnerValueType, Any]
) -> FutureResult[InnerValueType, Any]:
    result = flow(v, FutureResult.from_result)
    return result


def check_none_with_future_with_exception(
    exception: ExceptionType,
) -> Callable[[InnerValueType], FutureResult[InnerValueType, ExceptionType]]:
    def checkNone(
        v: InnerValueType,
    ) -> FutureResult[InnerValueType, ExceptionType]:
        return flow(
            v,
            Maybe.from_optional,
            lash(return_future_failure(exception)),
            bind(FutureResult.from_value),
        )

    return checkNone


def unwrap_maybe(v: Maybe[InnerValueType]) -> InnerValueType | None:
    return v.value_or(None)


def unwrap(v):
    try:
        return v.unwrap()
    except UnwrapFailedError as error:
        raise error.halted_container._inner_value


def assert_false():
    assert False


def assert_true():
    assert True


@curry
def assert_false_with_desc(desc):
    assert False, desc or ""


@curry
def assert_true_with_des(desc):
    assert True, desc or ""
    return True


@curry
def assert_equal(value, variable):
    assert variable == value


def throw_exception(error):
    match error:
        case Exception():
            raise error
        case _:
            assert False


@future_safe
async def assert_future_result_succesful(result: FutureResult):
    ioResult = await result.awaitable()
    assert is_successful(ioResult)
    result = flow(
        ioResult.unwrap(),
        unsafe_perform_io,
        # printResultWithText("assertFutureResultSuccesful result "),
    )
    return result


def raise_exception(ex: Exception):
    traceback.print_exc()
    raise ex


def pass_to(v: InnerValueType) -> Callable[[Any], InnerValueType]:
    return lambda _: v


def print_result_with_text(text: str):
    def printResult(v: InnerValueType) -> InnerValueType:
        print(f"{text} {v}")
        return v

    return printResult


def identity_factory(v: InnerValueType) -> Callable[[], InnerValueType]:
    def execute():
        return v

    return execute


def async_to_future_result(
    fnc: Callable[..., Coroutine[InnerValueType, Any, Any]],
) -> Callable[..., FutureResult[InnerValueType, Any]]:
    @future_safe
    async def execute(*args, **kwargs):
        result = await fnc(*args, **kwargs)
        return result

    return execute


def return_future_failure(
    e: ExceptionType,
) -> Callable[[Any], FutureResultE[ExceptionType]]:
    return lambda _: FutureResult.from_failure(e)


def unwrap_future_result_io(result: IOResult[InnerValueType, Any]) -> InnerValueType:
    return unsafe_perform_io(unwrap(result))


_ValueType = TypeVar("_ValueType", covariant=True)


async def unwrap_future_result(
    future_result: FutureResult[_ValueType, Any]
) -> _ValueType:
    io_result = await future_result.awaitable()
    match io_result:
        case IOSuccess(Success(data)):
            return data
        case IOSuccess(Failure(e)):
            raise e
        case IOFailure(Failure(e)):
            raise e
        case _:
            raise MainException("UNKNOW_PATTERN")


def collect_container(init):
    def with_items(items):
        return Fold.collect(items, init)

    return with_items


def filter_not_none(items: List[InnerValueType | None]) -> List[InnerValueType]:
    return list(filter(lambda item: item is not None, items))


def assert_not_none(v):
    assert v is not None


def assert_state_true(v):
    assert v


def feed_args(func):
    def execute(args: List[Any]):
        return func(*args)

    return execute


def feed_kwargs(func):
    def execute(kwargs: dict):
        return func(**kwargs)

    return execute


def throw_future_failed(exception: Exception):
    return lambda _: FutureFailure(exception)


def print_exception_with_traceback(exception: Exception):
    return traceback.print_tb(exception.__traceback__)


def not_nothing_or_throw_future_failed(
    exception: ExceptionType,
) -> Callable[[Maybe[InnerValueType]], FutureResult[InnerValueType, ExceptionType]]:
    def execute(
        value: Maybe[InnerValueType],
    ) -> FutureResult[InnerValueType, ExceptionType]:
        return flow(
            value,
            lash(throw_future_failed(exception)),
            bind(lambda v: FutureResult.from_value(v)),
        )

    return execute


def unwrap_future_io_maybe(
    value: IOResult[Maybe[InnerValueType], Any]
) -> InnerValueType | None:
    return unwrap_maybe(unsafe_perform_io(unwrap(value)))


def return_v(v: InnerValueType) -> Callable[[Any], InnerValueType]:
    return lambda _: v


FunctionType = TypeVar("FunctionType", bound=Callable)

P = ParamSpec("P")
T = TypeVar("T")


def apply(
    f: Callable[P, T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> Callable[[Any], T]:
    return lambda _: f(*args, **kwargs)


def unsafe_panic(f: Callable[..., Result | FutureResult]):
    def wrapper(*args, **kwargs):
        result = unwrap(f(*args, **kwargs))
        return result

    return wrapper


def execute(fn: Optional[Callable[..., T]], *args, **kwargs) -> T | None:
    return Maybe.from_optional(fn).map(lambda f: f(*args, **kwargs)).value_or(None)


def async_execute(fn: Callable[..., FutureResult]):
    def _(*args, **kwargs):
        return asyncio.run(fn(*args, **kwargs).awaitable())

    return _


def tap_excute_future(
    fn: Callable[[T], FutureResult]
) -> Callable[[T], FutureResult[T, Any]]:
    def execute(input: T) -> FutureResult[T, Any]:
        return fn(input).map(return_v(input))

    return execute


def tap_failure_execute_future(
    fn: Callable[[T], FutureResultE]
) -> Callable[[T], FutureResultE]:
    def execute(input: T) -> FutureResultE:
        return fn(input).bind(return_v(FutureFailure(input)))

    return execute


def must_be_true(
    exception: Optional[MainException] = None,
) -> Callable[[bool], Result]:
    def assert_true(value: bool) -> Result:
        has_value_and_exception = value is False and exception is not None
        match has_value_and_exception:
            case True:
                return Failure(exception)
            case False:
                return Success(value)

    return assert_true


def result_to_future_callable(fn: Callable[..., Result]):
    return lambda *args, **kwargs: FutureResult.from_result(fn(*args, **kwargs))


def with_default_value(original: Optional[T], defaultValue: T) -> T:
    return original or defaultValue



import base64


def censored_text(a_text: str):
    if a_text is None or len(a_text.strip()) == 0:
        return a_text

    return a_text[0:2] + "******" + a_text[-3:]


def parse_num_or_keeping(v: str):
    try:
        num = float(v)
        if num.is_integer():
            num = int(v)
        return num
    except Exception:
        return v


def split(v: str, sep: str):
    return v.split(sep)


def base64_to_string(b: bytes):
    return base64.b64decode(b).decode("utf-8")



"""
I am message producer, which facilitates sending messages to a BrokerChannel
A BrokerChannel may be either an Exchange or a Queue
"""


from typing import Union

from multimethod import multimethod
from pika import BasicProperties, DeliveryMode
from returns.maybe import Maybe, Nothing, Some
from returns.result import Failure, Result, Success, safe

from dino_seedwork_be.adapters.logger.SimpleLogger import SIMPLE_LOGGER
from dino_seedwork_be.logic.assertion_concern import AssertionConcern

from .exceptions import MessageException
from .Exchange import Exchange
from .MessageParameters import MessageParameters

# __all__ = ["MessageProducer"]


class MessageProducer(AssertionConcern):
    _exchange: Exchange

    def __init__(self, an_exchange: Exchange) -> None:
        super().__init__()
        self._exchange = an_exchange

    @staticmethod
    def factory(an_exchange: Exchange) -> "MessageProducer":
        return MessageProducer(an_exchange)

    def is_ready_for_publish(self) -> bool:
        return self.exchange().is_exchange_ready()

    def exchange(self) -> Exchange:
        return self._exchange

    @multimethod
    def send(self, a_message: Union[str, bytes], a_routing_key: str = "") -> Result:
        """
        Answers the receiver after sending a_message to my channel.
        This is a producer ignorance way to use either an exchange or
        a queue channel without requiring it to pass specific parameters.
        By answering myself I allow for sending message bursts.
        @param a_message the String text message to send
        @return MessageProducer
        """
        try:
            self.exchange().channel().unwrap().basic_publish(
                self.exchange().exchange_name().value_or(""),
                a_routing_key,
                properties=self.text_durability().value_or(None),
                body=a_message,
            )
            return Success(None)
        except Exception as e:
            SIMPLE_LOGGER.error("Error at basic publish a event: %s", e)
            return Failure(
                MessageException(
                    code="SEND_MESSAGE_FAILED",
                    message="Failed to message to channel " + str(e),
                )
            )

    @send.register
    def _(
        self,
        a_message_parameters: MessageParameters,
        a_message: Union[str, bytes],
        a_routing_key: str = "",
    ):
        """
        Answers the receiver after sending a_message to my channel
        with a_message_parameters as the message basic properties.
        This is a producer ignorance way to use either an exchange or
        a queue channel without requiring it to pass specific parameters.
        By answering myself I allow for sending message bursts.
        @param a_message_parameters the MessageParameters
        @param a_message the String text message to send
        @return MessageProducer
        """
        return self._check(a_message_parameters).bind(
            safe(
                lambda _: self.exchange()
                .channel()
                .unwrap()
                .basic_publish(
                    self.exchange().exchange_name().value_or(""),
                    a_routing_key,
                    properties=a_message_parameters.properties(),
                    body=a_message,
                )
            )
        )

    @send.register
    def _(
        self,
        an_exchange: str,
        a_routing_key: str,
        a_message_parameters: MessageParameters,
        a_message: Union[str, bytes],
    ) -> Result:
        """
        Answers the receiver after sending a_message to my channel
        with a_routing_key, a_message_parameters as the message basic properties.
        This is a producer ignorance way to use either an exchange or
        a queue channel without requiring it to pass specific parameters.
        By answering myself I allow for sending message bursts.
        @param an_exchange the String name of the exchange
        @param a_routing_key the String routing key
        @param a_message_parameters the MessageParameters
        @param a_message the String text message to send
        @return MessageProducer
        """

        return self._check(a_message_parameters).bind(
            safe(
                lambda _: self.exchange()
                .channel()
                .unwrap()
                .basic_publish(
                    an_exchange,
                    a_routing_key,
                    properties=a_message_parameters.properties(),
                    body=a_message,
                )
            )
        )

    def binary_durability(self) -> Maybe[BasicProperties]:
        match self.exchange().is_durable():
            case True:
                return Some(
                    BasicProperties(
                        content_type="application/octet-stream",
                        delivery_mode=DeliveryMode.Persistent,
                        priority=0,
                    ),
                )
            case False:
                return Nothing

    def text_durability(self) -> Maybe[BasicProperties]:
        match self.exchange().is_durable():
            case True:
                return Some(
                    BasicProperties(
                        content_type="text/plain",
                        delivery_mode=DeliveryMode.Transient,
                        priority=0,
                    )
                )
            case False:
                return Nothing

    def _check(self, a_message_parameters) -> Result:
        return self.assert_state_false(
            self.exchange().is_durable() != a_message_parameters.is_durable(),
            code=Some("MSG_PARAMETER_DURABLE_NOT_MATCH_WITH_BROKER"),
            a_message=Some(
                "message parameter durable property is not match with broker component durable"
            ),
        )

    def run(self):
        self.exchange().run()

    def close(self):
        """
        Closes me, which closes my broker channel.
        """
        self.exchange().close()



from typing import List, Optional, TypeVar

from faker import Faker
from returns.pipeline import flow

fake = Faker()
Faker.seed(0)  # This will raise a TypeError

ElementType = TypeVar("ElementType")


def random_elements_or_none(elements: List[ElementType]) -> Optional[List[ElementType]]:
    return flow(
        tuple(elements),
        lambda elements: fake.random_elements(elements=elements, unique=True),
        lambda v: (v, None),
        fake.random_element,
    )


def random_element_or_none(element: ElementType) -> Optional[ElementType]:
    return fake.random_element(elements=(element, None))



from .AlchemyQuerier import AlchemyQuerier as AlchemyQuerier
from .BaseModel import UUIDBaseModel as UUIDBaseModel
from .connection import alchemy_session_factory as alchemy_session_factory
from .connection import engine_factory as engine_factory
from .Repository import StandardAlchemyRepository as StandardAlchemyRepository
from .SerializableBase import SerializableEntity as SerializableEntity
from .uow import SqlAlchemyUnitOfWork as SqlAlchemyUnitOfWork
from .util.alchemy_sql import (alchemy_execute_query_on_repository,
                               tuple_row_to_dict, tuple_rows_to_dict)

__all__ = [
    "AlchemyQuerier",
    "UUIDBaseModel",
    "alchemy_execute_query_on_repository",
    "alchemy_session_factory",
    "engine_factory",
    "tuple_row_to_dict",
    "tuple_rows_to_dict",
    "StandardAlchemyRepository",
    "SerializableEntity",
    "SqlAlchemyUnitOfWork",
]



import json
from typing import Generic, TypeVar

import jsonpickle
from returns.result import Failure, Result, Success

from dino_seedwork_be.serializer.AbstractSerializer import AbstractSerializer

from .Notification import Notification

NotificationT = TypeVar("NotificationT", bound=Notification)

__all__ = ["NotificationSerializer"]


class NotificationSerializer(
    AbstractSerializer,
    Generic[NotificationT],
):
    ins: "NotificationSerializer"

    @staticmethod
    def instance() -> "NotificationSerializer":
        return NotificationSerializer.ins

    def serialize(self, aNotification: NotificationT) -> Result[str, Exception]:
        try:
            return Success(
                "%s" % str(jsonpickle.encode(aNotification, unpicklable=False))
            )
        except Exception as error:
            return Failure(error)

    def deserialize(self, aJson) -> Result[Notification, Exception]:
        try:
            return Success(Notification.restore(json.loads(aJson)))
        except Exception as error:
            return Failure(error)


NotificationSerializer.ins = NotificationSerializer()



from abc import abstractmethod
from datetime import datetime

from multimethod import multimethod
from returns.future import FutureFailure, FutureResult

from dino_seedwork_be.exceptions import MainException

from .MessageType import MessageType

# __all__ = ["MessageListener"]


class MessageListener:
    """
    * I am a message listener, which is given each message received
    * by a MessageConsumer. I am also an adapter because I provide
    * defaults for both handleMessage() behaviors. A typical subclass
    * would override one or the other handleMessage() based on its
    * type and leave the remaining handleMessage() defaulted since
    * it will never be used by MessageConsumer.
    *
    """

    # My type, which indicates whether I listen for BINARY or TEXT messages.
    _type: MessageType

    def __init__(self, a_type: MessageType) -> None:
        self.set_type(a_type)

    def type(self) -> MessageType:
        """
        * Answers my type.
        * @return Type
        """
        return self._type

    def set_type(self, a_type: MessageType):
        """
        * Sets my type.
        * @param a_type the Type to set as my type
        """
        self._type = a_type

    @abstractmethod
    @multimethod
    def handle_message(
        self,
        a_type: str,
        a_message_id: str,
        a_time_stamp: datetime,
        a_binary_message: bytes,
        a_delivery_tag: int,
        is_redelivery: bool,
    ) -> FutureResult:

        """
        * Handles a_binary_message. If any MessageException is thrown by
        * my implementor its is_retry() is examined and, if true, the
        * message being handled will be nack'd and re-queued. Otherwise,
        * if its is_retry() is false the message will be rejected/failed
        * (not re-queued). If any other Exception is thrown the message
        * will be considered not handled and is rejected/failed.
        * @param a_type the String type of the message if sent, or null
        * @param a_message_id the String id of the message if sent, or null
        * @param a_timestamp the Date timestamp of the message if sent, or null
        * @param a_binary_message the byte[] containing the binary message
        * @param a_delivery_tag the long tag delivered with the message
        * @param is_redelivery the boolean indicating whether or not this message is a redelivery
        * @throws Exception when any problem occurs and the message must not be acknowledged
        """
        return FutureFailure(
            MainException(
                code="METHOD_NOT_IMPLEMENTED",
                message="Handle message method must be implemented",
            )
        )

    @abstractmethod
    @handle_message.register
    def _(
        self,
        a_type: str,
        a_message_id: str,
        a_time_stamp: datetime,
        a_text_message: str,
        a_delivery_tag: int,
        is_redelivery: bool,
    ):
        """
        * Handles a_text_message. If any MessageException is thrown by
        * my implementor its is_retry() is examined and, if true, the
        * message being handled will be nack'd and re-queued. Otherwise,
        * if its is_retry() is false the message will be rejected/failed
        * (not re-queued). If any other Exception is thrown the message
        * will be considered not handled and is rejected/failed.
        * @param a_type the String type of the message if sent, or null
        * @param a_message_id the String id of the message if sent, or null
        * @param a_timestamp the Date timestamp of the message if sent, or null
        * @param a_text_message the byte[] containing the binary message
        * @param a_delivery_tag the long tag delivered with the message
        * @param is_redelivery the boolean indicating whether or not this message is a redelivery
        * @throws Exception when any problem occurs and the message must not be acknowledged
        """
        return FutureFailure(
            MainException(
                code="METHOD_NOT_IMPLEMENTED",
                message="Handle message method must be implemented",
            )
        )



from pika import BasicProperties, DeliveryMode

# __all__ = ["MessageParameters"]


class MessageParameters:
    _properties: BasicProperties

    def __init__(self, a_properties: BasicProperties) -> None:
        super().__init__()
        self.set_properties(a_properties)

    def properties(self) -> BasicProperties:
        return self._properties

    def set_properties(self, a_properties: BasicProperties):
        self._properties = a_properties

    def is_durable(self) -> bool:
        return self.properties().delivery_mode == 2

    @staticmethod
    def durable_text_parameters(a_type: str, a_message_id: str, a_timestamp: int):
        properties = BasicProperties(
            content_type="text/plain",
            content_encoding=None,
            headers=None,
            delivery_mode=DeliveryMode.Persistent,
            priority=0,
            correlation_id=None,
            reply_to=None,
            expiration=None,
            message_id=a_message_id,
            timestamp=a_timestamp,
            type=a_type,
            user_id=None,
            app_id=None,
            cluster_id=None,
        )
        return MessageParameters(properties)

    @staticmethod
    def text_parameters(a_type: str, a_message_id: str, a_timestamp: int):
        properties = BasicProperties(
            content_type="text/plain",
            content_encoding=None,
            headers=None,
            delivery_mode=DeliveryMode.Transient,
            priority=0,
            correlation_id=None,
            reply_to=None,
            expiration=None,
            message_id=a_message_id,
            timestamp=a_timestamp,
            type=a_type,
            user_id=None,
            app_id=None,
            cluster_id=None,
        )
        return MessageParameters(properties)



from datetime import datetime
from uuid import uuid4

from sqlalchemy import Column, DateTime
from sqlalchemy.dialects.postgresql import UUID

__all__ = ["UUIDBaseModel"]


class UUIDBaseModel:
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid4)
    created_at = Column(DateTime(timezone=True), default=datetime.now())
    updated_at = Column(DateTime(timezone=True))



from abc import ABC, abstractmethod


class AbstractLogger(ABC):

    _tag: str

    def __init__(self, tag: str):
        self.set_tag(tag)

    def set_tag(self, a_tag: str):
        self._tag = a_tag

    def tag(self) -> str:
        return self._tag

    @abstractmethod
    def info(self, message: str):
        pass

    @abstractmethod
    def error(self, message: str):
        pass

    @abstractmethod
    def warning(self, message: str):
        pass



from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum

from .ProcessId import ProcessId


class ProcessCompletionType(Enum):
    NOT_COMPLETED = "NOT_COMPLETED"
    COMPLETED_NORMALLY = "COMPLETED_NORMALLY"
    TIMED_OUT = "TIMED_OUT"


class IProcess(ABC):
    @abstractmethod
    def allowable_duration() -> int:
        ...

    @abstractmethod
    def can_timeout() -> bool:
        ...

    @abstractmethod
    def current_duration() -> int:
        ...

    @abstractmethod
    def description() -> str:
        ...

    @abstractmethod
    def did_processing_complete() -> bool:
        ...

    @abstractmethod
    def inform_timeout_date(self, a_time_out_date: datetime):
        ...

    @abstractmethod
    def is_completed() -> bool:
        ...

    @abstractmethod
    def is_timeout() -> bool:
        ...

    def is_not_completed(self) -> bool:
        return not self.is_completed()

    @abstractmethod
    def process_completion_type(self) -> ProcessCompletionType:
        ...

    def process_id(self) -> ProcessId:
        ...

    def start_time(self) -> datetime:
        ...

    def time_constrained_process_tracker(self):
        ...



from abc import abstractmethod
from typing import List

from returns.curry import partial
from returns.future import FutureResult, FutureSuccess
from returns.iterables import Fold
from returns.pipeline import flow, pipe
from returns.pointfree import bind

from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import \
    DBSessionUser
from dino_seedwork_be.application import ApplicationLifeCycleUsecase

from .TimeConstrainedProcessTrackerRepository import \
    TimeConstrainedProcessTrackerRepository


class AbstractProcessApplicationService(ApplicationLifeCycleUsecase):

    _process_tracker_repository: TimeConstrainedProcessTrackerRepository

    def process_tracker_repository(self) -> TimeConstrainedProcessTrackerRepository:
        return self._process_tracker_repository

    def __init__(
        self, a_processor_tracker_repository: TimeConstrainedProcessTrackerRepository
    ) -> None:
        self._process_tracker_repository = a_processor_tracker_repository
        super().__init__()

    def get_session_users(self) -> List[DBSessionUser]:
        return [self.process_tracker_repository()]

    @abstractmethod
    def check_for_timedout_processes(self):
        pass

    def _execute_check_for_timedout_process(self) -> FutureResult:
        return flow(
            self.process_tracker_repository().all_timed_out(),
            bind(
                pipe(
                    partial(
                        map,
                        lambda tracker: flow(
                            tracker.inform_process_timedout(
                                self.process_tracker_repository().timeout_event_factory()
                            ),
                            bind(
                                lambda _: self.process_tracker_repository().save(
                                    tracker
                                )
                            ),
                        ),
                    ),
                    lambda results: Fold.collect_all(results, FutureSuccess(())),
                )
            ),
        )



from .AbstractDomainEventDict import AbstractDomainEventDict
from .AggregateRoot import AggregateRoot
from .DomainAssertionConcern import DomainAssertionConcern
from .DomainEvent import DomainEvent, EmptyProps
from .DomainEventPublisher import DomainEventPublisher
from .DomainEventSubscriber import DomainEventSubscriber
from .DomainService import DomainService
from .Entity import BaseRawAttributes, Entity
from .event.EventSerializer import EventSerializer
from .event.EventStore import EventStore
from .event.EventStoreSubscriber import EventStoreSubscriber
from .event.StoredEvent import StoredEvent
from .exceptions import (BusinessRuleValidationException, DomainException,
                         DomainIllegalArgumentException,
                         DomainIllegalStateException)
from .IdentifiedDomainObject import IdentifiedDomainObject
from .IdentifiedValueObject import IdentifiedValueObject
from .mixins import OrderItemMixin
from .utils import (exception_to_domain_exception, get_identity,
                    get_raw_identity)
from .value_object.AbstractIdentity import AbstractIdentity
from .value_object.AbstractValueObject import ValueObject
from .value_object.File import File
from .value_object.FullName import FullName
from .value_object.ImageURL import ImageURL
from .value_object.NID import NID
from .value_object.RegexValue import StringWithRegex
from .value_object.URL import URL
from .value_object.UUID import UUID

__all__ = [
    "AbstractIdentity",
    "ValueObject",
    "File",
    "FullName",
    "ImageURL",
    "NID",
    "StringWithRegex",
    "URL",
    "UUID",
    "EventStoreSubscriber",
    "EventSerializer",
    "EventStore",
    "StoredEvent",
    "AbstractDomainEventDict",
    "AggregateRoot",
    "DomainAssertionConcern",
    "DomainEvent",
    "EmptyProps",
    "DomainEventPublisher",
    "DomainEventSubscriber",
    "DomainService",
    "BaseRawAttributes",
    "Entity",
    "BusinessRuleValidationException",
    "DomainException",
    "DomainIllegalArgumentException",
    "DomainIllegalStateException",
    "IdentifiedDomainObject",
    "IdentifiedValueObject",
    "OrderItemMixin",
    "exception_to_domain_exception",
    "get_identity",
    "get_raw_identity",
]



from datetime import datetime
from typing import Any, List, Type

from returns.curry import partial
from returns.functions import tap
from returns.future import FutureResult, FutureSuccess, future_safe
from returns.pipeline import flow
from returns.pointfree import alt, bind, map_
from sqlalchemy import Column, DateTime, Integer, String, select
from sqlalchemy.sql.functions import count

from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import \
    DBSessionUser
from dino_seedwork_be.domain.DomainEvent import DomainEvent
from dino_seedwork_be.domain.event.EventSerializer import EventSerializer
from dino_seedwork_be.domain.event.EventStore import EventStore
from dino_seedwork_be.domain.event.StoredEvent import StoredEvent
from dino_seedwork_be.utils import (async_to_future_result, feed_kwargs,
                                    print_result_with_text, return_v)

# __all__ = ["SqlAlchemyBaseEvent", "SqlAlchemyEventStore"]


class SqlAlchemyBaseEvent:
    __tablename__ = "event_store"
    id = Column(Integer(), primary_key=True, autoincrement=True)
    occurred_on = Column(DateTime(timezone=True), default=datetime.now())
    body = Column(String())
    type_name = Column(String(), nullable=False)


class SqlAlchemyEventStore(EventStore, DBSessionUser):
    _db_model: Type[SqlAlchemyBaseEvent]

    def __init__(self, db_model: Type[SqlAlchemyBaseEvent]) -> None:
        self._db_model = db_model

    @future_safe
    async def all_stored_events_since(self, an_event_id: int) -> List[StoredEvent]:
        stmt = select(self._db_model).where(self._db_model.id > an_event_id)
        result = (await self.session().execute(stmt)).scalars().all()
        return flow(
            result,
            partial(
                map,
                lambda r: StoredEvent(
                    id=r.id,
                    occurred_on=r.occurred_on,
                    type_name=r.type_name,
                    body=r.body,
                ),
            ),
            list,
        )

    def all_stored_events_between(
        self, a_low_stored_event_id: int, a_high_stored_event_id: int
    ) -> FutureResult[List[StoredEvent], Any]:
        return FutureSuccess([])

    def append(self, an_domain_event: DomainEvent) -> FutureResult[StoredEvent, Any]:
        stored_event_result = flow(
            an_domain_event,
            EventSerializer.instance().serialize,
            map_(
                lambda body: {
                    "body": body,
                    "occurred_on": an_domain_event.occurred_on(),
                    "type_name": an_domain_event.type(),
                }
            ),
            map_(feed_kwargs(StoredEvent)),
        )
        return flow(
            stored_event_result.map(
                lambda v: {
                    "occurred_on": v.occurred_on(),
                    "type_name": v.type_name(),
                    "body": v.body(),
                }
            ),
            map_(feed_kwargs(self._db_model)),
            map_(tap(self.session().add)),
            bind(lambda _: async_to_future_result(self.session().commit)()),
            map_(print_result_with_text("Commit event to event store")),
            bind(return_v(FutureResult.from_result(stored_event_result))),
            alt(tap(lambda _: async_to_future_result(self.session().rollback)())),
        )

    def close(self):
        ...

    @future_safe
    async def count_events(self) -> int:
        stmt = select(count(self._db_model.id))
        return (await self.session().execute(stmt)).scalar()



from typing import Any, List

from dino_seedwork_be.implementation.adapter.storage.sql.alchemy.Repository import \
    StandardAlchemyRepository

# __all__ = [
#     "alchemy_execute_query_on_repository",
#     "tuple_row_to_dict",
#     "tuple_rows_to_dict",
# ]


def alchemy_execute_query_on_repository(repository: StandardAlchemyRepository):
    async def proxy(*args, **kwargs):
        return await repository.session().execute(*args, **kwargs)

    return proxy


def tuple_row_to_dict(row: Any):
    return dict(zip(row.keys(), row))


def tuple_rows_to_dict(rows: List[Any]):
    return [tuple_row_to_dict(row) for row in rows]



import json
from datetime import datetime
from typing import Optional, TypedDict

from returns.functions import tap
from returns.maybe import Maybe, Some
from returns.pipeline import flow
from returns.pointfree import map_
from returns.result import Result, Success, safe
from toolz.dicttoolz import get_in

from dino_seedwork_be.logic.assertion_concern import AssertionConcern
from dino_seedwork_be.serializer.Serializable import JSONSerializable
from dino_seedwork_be.utils.date import now_utc, to_iso_format
from dino_seedwork_be.utils.functional import unwrap


class EmptyProps(TypedDict):
    pass


class DomainEvent(AssertionConcern, JSONSerializable):
    _id: int | None
    _version: int = 0
    _occurred_on: datetime
    _name: str
    _props: dict = {}

    def __init__(
        self,
        name: str,
        id: int | None = None,
        version: int = 0,
        occurred_on: Optional[datetime] = None,
        props: dict = {},
    ):
        unwrap(
            flow(
                None,
                lambda _: self.set_version(version),
                map_(lambda _: self.set_occurred_on(occurred_on or now_utc())),
                map_(lambda _: self.set_name(name)),
                map_(lambda _: self.set_props(props)),
                map_(lambda _: self.set_id(id)),
            )
        )

    def __eq__(self, __o: object) -> bool:
        match __o:
            case DomainEvent() as domain_event:
                return (
                    domain_event.name() == self.name()
                    and domain_event.props() == self.props()
                )
            case _:
                return False

    @staticmethod
    @safe
    def factory(
        occurred_on: datetime,
        name: str,
        version: int = 0,
        props: dict = {},
    ):
        return DomainEvent(name, None, version, occurred_on, props)

    def occurred_on(self) -> datetime:
        return self._occurred_on

    def version(self) -> int:
        return self._version

    def props(self) -> dict:
        return self._props

    def set_occurred_on(self, a_datetime: datetime) -> Result:
        self._occurred_on = a_datetime
        return Success(None)

    def set_version(self, a_version: int) -> Result:
        self._version = a_version
        return Success(None)

    def set_id(self, id: int | None) -> Result:
        self._id = id
        return Success(None)

    def set_name(self, a_name: str) -> Result:
        def set(_):
            self._name = a_name

        return flow(
            self.assert_argument_not_empty(Some(a_name)),
            map_(tap(set)),
        )

    def set_props(self, props: dict) -> Result:
        self._props = props
        return Success(None)

    def get_prop_attr(self, attr_name: str):
        return self._props[attr_name]

    def name(self) -> str:
        return self._name

    def type(self) -> str:
        return self.name()

    def body_json(self) -> str:
        return json.dumps(self.props(), default=str) or "{}"

    def props_to_dict(self):
        return self._props

    def as_dict(self):
        return {
            "version": self.version(),
            "occurred_on": to_iso_format(self.occurred_on()),
            "name": self.name(),
            "props": self.props_to_dict(),
            "id": self._id,
        }

    def id(self) -> Maybe[int]:
        return Maybe.from_optional(self._id)

    @staticmethod
    def restore(a_dict):
        return DomainEvent(
            version=get_in(["version"], a_dict, 0),
            occurred_on=Maybe.from_optional(get_in(["occurred_on"], a_dict, None))
            .map(str)
            .map(datetime.fromisoformat)
            .value_or(None),
            name=str(get_in(["name"], a_dict)),
            props=dict(get_in(["props"], a_dict, {})),
            id=Maybe.from_optional(get_in(["id"], a_dict))
            .map(lambda v: int(v))
            .value_or(None),
        )



from abc import ABC, abstractmethod

from returns.future import FutureResult

__all__ = ["UserNotificationPushDrivingAdapter"]


class UserNotificationPushDrivingAdapter(ABC):
    @abstractmethod
    def push_to_device(
        self, content: dict, type: str, device_token: str
    ) -> FutureResult:
        pass



from .adapters import *
from .application import *
from .domain import *
from .exceptions import (ExceptionCode, IllegalArgumentException,
                         IllegalStateException, MainException,
                         NotImplementError, except_locs)
from .fp import *
from .logic import *
from .media import *
from .process import *
from .repository import *
from .serializer import *
from .types import *
from .utils import *

__all__ = [
    "NotImplementError",
    "ExceptionCode",
    "MainException",
    "IllegalStateException",
    "IllegalArgumentException",
    "except_locs",
]



from .SqlAlchemyEventStore import SqlAlchemyBaseEvent, SqlAlchemyEventStore
from .SqlAlchemyPublishedNotificationTrackerStore import (
    SqlAlchemyBasePublishedNotifTracker,
    SqlAlchemyPublishedNotificationTrackerStore)

__all__ = [
    "SqlAlchemyPublishedNotificationTrackerStore",
    "SqlAlchemyEventStore",
    "SqlAlchemyBaseEvent",
    "SqlAlchemyBasePublishedNotifTracker",
]



from typing import Any

from returns.future import FutureResult

from dino_seedwork_be.adapters.messaging.notification.EventHandlingTracker import \
    EventHandlingTracker
from dino_seedwork_be.adapters.persistance.key_value.AbstractKeyValueRepository import \
    AbstractKeyValueRepository
from dino_seedwork_be.utils.params import cast_bool_from_str


class KeyValueEventHandlingTracker(EventHandlingTracker):
    _key_value_repository: AbstractKeyValueRepository
    _prefix: str

    def __init__(
        self,
        key_value_repository: AbstractKeyValueRepository,
        prefix: str = "notification_tracker",
    ) -> None:
        self._prefix = prefix
        self._key_value_repository = key_value_repository

    def prefix(self) -> str:
        return self._prefix

    def _key_with_prefix(self, key: str) -> str:
        return f"{self.prefix()}_{key}"

    def check_if_notif_handled(self, a_message_id: str) -> FutureResult[bool, Any]:
        return self._key_value_repository.get(self._key_with_prefix(a_message_id)).map(
            lambda maybe_result: maybe_result.map(cast_bool_from_str).value_or(False)
        )

    def mark_notif_as_handled(self, a_message_id: str) -> FutureResult:
        return self._key_value_repository.set(
            self._key_with_prefix(a_message_id), "True"
        )

    def unmark_notif_as_handled(self, a_message_id: str) -> FutureResult:
        return self._key_value_repository.set(
            self._key_with_prefix(a_message_id), "False"
        )



from .date import now_utc, to_iso_format

__all__ = ["to_iso_format", "now_utc"]

from .dict import dict_to_cls, extract, keys, values

__all__.extend(["keys", "values", "dict_to_cls", "extract"])

from .faker_helpers import random_element_or_none, random_elements_or_none

__all__.extend(["random_element_or_none", "random_elements_or_none"])

from .functional import (apply, assert_equal, assert_false,
                         assert_false_with_desc,
                         assert_future_result_succesful, assert_not_none,
                         assert_state_true, assert_true, assert_true_with_des,
                         async_to_future_result,
                         check_none_with_future_with_exception,
                         collect_container, execute, feed_args, feed_identity,
                         feed_kwargs, filter_not_none, for_each,
                         identity_factory, map_to_list, maybe_to_future,
                         must_be_true, not_nothing_or_throw_future_failed,
                         pass_to, print_exception_with_traceback,
                         print_result_with_text, raise_exception,
                         result_to_future, result_to_future_callable,
                         return_future_failure, return_v, set_private_attr,
                         set_protected_attr, set_public_attr,
                         tap_excute_future, tap_failure_execute_future,
                         throw_exception, throw_future_failed, unsafe_panic,
                         unwrap, unwrap_future_io_maybe,
                         unwrap_future_result_io, unwrap_maybe,
                         with_default_value)

__all__.extend(
    [
        "apply",
        "assert_equal",
        "assert_false",
        "assert_false_with_desc",
        "assert_future_result_succesful",
        "assert_not_none",
        "assert_state_true",
        "assert_true",
        "assert_true_with_des",
        "async_to_future_result",
        "check_none_with_future_with_exception",
        "collect_container",
        "execute",
        "feed_args",
        "feed_identity",
        "feed_kwargs",
        "filter_not_none",
        "for_each",
        "identity_factory",
        "map_to_list",
        "maybe_to_future",
        "not_nothing_or_throw_future_failed",
        "pass_to",
        "print_result_with_text",
        "print_exception_with_traceback",
        "raise_exception",
        "result_to_future",
        "return_future_failure",
        "set_private_attr",
        "set_protected_attr",
        "set_public_attr",
        "tap_excute_future",
        "tap_failure_execute_future",
        "throw_exception",
        "throw_future_failed",
        "unsafe_panic",
        "unwrap",
        "unwrap_future_io_maybe",
        "unwrap_future_result_io",
        "unwrap_maybe",
        "return_v",
        "must_be_true",
        "result_to_future_callable",
        "with_default_value",
    ]
)

from .image import get_image_dimension, get_image_file_size

__all__.extend(["get_image_dimension", "get_image_file_size"])

from .list import shallow_compare_list

__all__.extend(["shallow_compare_list"])

from .meta import get_class_name, get_local_classname

__all__.extend(["get_local_classname", "get_class_name"])

from .none_or_instance import none_or_instance, none_or_transform

__all__.extend(["none_or_instance", "none_or_transform"])

from .number import increase, is_in_range, negate

__all__.extend(["increase", "is_in_range", "negate"])

from .params import cast_bool_from_str, get_env, get_env_with, get_environment

__all__.extend(["cast_bool_from_str", "get_env_with", "get_env", "get_environment"])

from .process.ThreadLocal import ThreadLocal

__all__.extend(["ThreadLocal"])



import json

from returns.result import safe

from .AbstractSerializer import AbstractSerializer


class SimpleJSONSerializer(AbstractSerializer[dict]):
    @safe
    def serialize(self, obj: dict) -> str:
        return json.dumps(obj)

    @safe
    def deserialize(self, an_json: str) -> dict:
        return json.loads(an_json)



from typing import Optional, Union

from redis import Redis
from returns.future import future_safe
from returns.maybe import Maybe

from dino_seedwork_be.adapters.persistance.key_value.AbstractKeyValueRepository import \
    AbstractKeyValueRepository


class RedisPyRepository(AbstractKeyValueRepository):

    _redis: Redis

    def __init__(self, redis: Redis, prefix: str = "") -> None:
        self._redis = redis
        super().__init__(prefix)

    def redis(self):
        return self._redis

    def _key_with_prefix(self, key: str) -> str:
        prefix = self._prefix
        return f"{prefix}:{key}" if len(prefix) > 0 else key

    @future_safe
    async def set(
        self,
        key: str,
        value: Union[bytes, memoryview, str, int, float],
        expired_seconds: Optional[int] = None,
    ):
        return self.redis().set(
            name=self._key_with_prefix(key), value=value, ex=expired_seconds
        )

    @future_safe
    async def get(self, key: str) -> Maybe:
        return Maybe.from_optional(self.redis().get(name=self._key_with_prefix(key)))



from .AbstractSerializer import AbstractSerializer
from .Serializable import JSONSerializable
from .SimpleJSONSerializer import SimpleJSONSerializer

__all__ = ["AbstractSerializer", "JSONSerializable", "SimpleJSONSerializer"]



from returns.maybe import Maybe

from dino_seedwork_be.domain.Entity import Entity, RawAttributes
from dino_seedwork_be.domain.IdentifiedDomainObject import IdentityType
from dino_seedwork_be.utils.functional import print_exception_with_traceback

from .exceptions import DomainException

# __all__ = ["get_identity", "get_raw_identity"]


def get_identity(an_entity: Entity[RawAttributes, IdentityType]) -> Maybe[IdentityType]:
    return an_entity.identity()


def get_raw_identity(an_entity: Entity) -> Maybe[str]:
    return an_entity.id_as_string()


def exception_to_domain_exception(code: str, prefix: str, exception: Exception):
    return DomainException(
        code=code,
        message=f"[{prefix} - {str(exception)}] {print_exception_with_traceback(exception)}",
    )



from returns.maybe import Some

from dino_seedwork_be.domain.exceptions import DomainException
from dino_seedwork_be.domain.value_object.AbstractValueObject import \
    ValueObject

# __all__ = ["FullName"]


class FullName(ValueObject):
    _first_name: str | None = None
    _last_name: str | None = None

    def __init__(self, first_name: str | None, last_name: str | None):
        super().__init__()
        if first_name is not None:
            self._set_first_name(first_name)
        if last_name is not None:
            self._set_last_name(last_name)

    def __eq__(self, obj):
        if isinstance(obj, FullName):
            return (
                self.first_name() == obj.first_name()
                and self.last_name() == obj.last_name()
            )
        return False

    def first_name(self) -> str | None:
        return self._first_name

    def last_name(self) -> str | None:
        return self._last_name

    def as_formatted_name(self):
        return f"{self._last_name} {self._first_name}"

    def with_changed_first_name(self, aFirstName: str):
        return FullName(first_name=aFirstName, last_name=self._last_name)

    def with_changed_last_name(self, aSecondName: str):
        return FullName(last_name=aSecondName, first_name=self._first_name)

    def _set_first_name(self, a_value: str):
        try:
            self.assert_argument_not_empty(
                Some(a_value), Some("First name is required")
            ).unwrap()
            self.assert_argument_length(
                a_value,
                a_minimum=1,
                a_maximum=50,
                a_message=Some("First name must be 50 characters or less"),
            ).unwrap()
            self.assert_argument_regex(
                a_value,
                r"^[A-Z\s](?:[A-Za-z_ÀÁÂÃÈÉÊÌÍÒÓÔÕÙÚĂĐĨŨƠàáâãèéêìíòóôõùúăđĩũơƯĂẠẢẤẦẨẪẬẮẰẲẴẶẸẺẼỀỀỂưăạảấầẩẫậắằẳẵặẹẻẽềềểỄỆỈỊỌỎỐỒỔỖỘỚỜỞỠỢỤỦỨỪễếệỉịọỏốồổỗộớờởỡợụủứừỬỮỰỲỴÝỶỸửữựỳỵỷỹ\s])+$",
                Some(
                    "First name must be at least one character in length, \
                starting with a capital letter."
                ),
            ).unwrap()
            self._first_name = a_value
        except Exception as error:
            raise DomainException(code="FIRST_NAME_INVALID", message=str(error))

    def _set_last_name(self, a_value: str):
        try:
            self.assert_argument_not_empty(
                Some(a_value), Some("Last name is required")
            ).unwrap()
            self.assert_argument_length(
                a_value,
                a_minimum=1,
                a_maximum=50,
                a_message=Some("last name must be 50 characters or less"),
            ).unwrap()
            self.assert_argument_regex(
                a_value,
                r"^[A-Z\s](?:[A-Za-z_ÀÁÂÃÈÉÊÌÍÒÓÔÕÙÚĂĐĨŨƠàáâãèéêìíòóôõùúăđĩũơƯĂẠẢẤẦẨẪẬẮẰẲẴẶẸẺẼỀỀỂưăạảấầẩẫậắằẳẵặẹẻẽềềểỄỆỈỊỌỎỐỒỔỖỘỚỜỞỠỢỤỦỨỪễếệỉịọỏốồổỗộớờởỡợụủứừỬỮỰỲỴÝỶỸửữựỳỵỷỹ\s])+$",
                Some("Last name must be at least one character in length."),
            ).unwrap()
            self._last_name = a_value
        except Exception as error:
            raise DomainException(code="LAST_NAME_INVALID", message=str(error))



import uuid
from typing import Any, Optional

from returns.future import FutureResult, FutureSuccess
from returns.pipeline import flow, pipe
from returns.pointfree import bind, lash

from dino_seedwork_be.domain.value_object.UUID import UUID
from dino_seedwork_be.utils.functional import return_v

from .IRepository import EntityType, Repository


class RepositoryUUID(Repository[EntityType]):
    def get_next_id(self, simple: Optional[bool] = False) -> FutureResult[UUID, Any]:
        def check_exist_and_gen() -> FutureResult:
            next_id_candidate = UUID(uuid.uuid4())
            return flow(
                next_id_candidate,
                self.get_by_id,
                bind(
                    pipe(
                        bind(return_v(FutureSuccess(next_id_candidate))),
                        lash(lambda _: check_exist_and_gen()),
                    )
                ),
            )

        next_id_candidate = uuid.uuid4()
        match simple:
            case True:
                return FutureSuccess(UUID(next_id_candidate))
            case False:
                return check_exist_and_gen()

        return FutureSuccess(UUID(next_id_candidate))



import datetime
from abc import abstractmethod
from typing import Generic, TypedDict, TypeVar, cast

from returns.maybe import Maybe, Nothing, Some
from returns.result import Result, Success, safe
from typing_extensions import Self

from dino_seedwork_be.domain.exceptions import DomainException
from dino_seedwork_be.utils.date import now_utc
from dino_seedwork_be.utils.functional import unwrap

from .IdentifiedDomainObject import IdentifiedDomainObject, IdentityType


class BaseRawAttributes(TypedDict):
    created_at: datetime.datetime | None
    updated_at: datetime.datetime | None


RawAttributes = TypeVar("RawAttributes", bound=BaseRawAttributes)


# __all__ = ["Entity", "BaseRawAttributes"]


class Entity(
    Generic[RawAttributes, IdentityType], IdentifiedDomainObject[IdentityType]
):
    _created_at: Maybe[datetime.datetime] = Nothing
    _updated_at: Maybe[datetime.datetime] = Nothing

    _concurrency_version: int = 0

    def __init__(self, id: Maybe[IdentityType] = Nothing):
        self._concurrency_version = 0
        super().__init__(id)

    def updated_at(self) -> Maybe:
        return self._updated_at

    def created_at(self) -> Maybe:
        return self._created_at

    def set_update_at(self, a_date_time: datetime.datetime) -> Result:
        self._updated_at = Some(a_date_time)
        return Success("OK")

    def set_created_at(self, a_date_time: datetime.datetime) -> Result:
        self._created_at = Some(a_date_time)
        return Success("OK")

    def concurrency_version(self) -> int:
        return self._concurrency_version

    def increase_concurrency_version(self):
        self._concurrency_version += 1

    def __eq__(self, __o) -> bool:
        if isinstance(__o, Entity):
            return self.identity() == __o.identity()
        else:
            return False

    @classmethod
    @safe(exceptions=(DomainException,))
    def create(cls, raw_attributes: RawAttributes, id: Maybe[IdentityType]) -> Self:
        entity = cls(id)
        match raw_attributes["created_at"]:
            case datetime.datetime() as created_at:
                unwrap(entity.set_created_at(created_at))
            case None:
                unwrap(entity.set_created_at(now_utc()))

        match raw_attributes["updated_at"]:
            case datetime.datetime() as updated_at:
                unwrap(entity.set_update_at(updated_at))
        unwrap(entity.init_by_atributes(raw_attributes))
        return entity

    @abstractmethod
    def init_by_atributes(self, raw_attributes: RawAttributes) -> Result:
        pass



from .BrokerComponent import BrokerComponent as RabbitMQBrokerComponent
from .ConnectionSettings import \
    ConnectionSettings as RabbitMQConnectionSettings
from .exceptions import MessageException as MessageException
from .Exchange import Exchange as RabbitMQExchange
from .ExchangeListener import ExchangeListener as RabbitMQExchangeListener
from .MessageConsumer import MessageConsumer as RabbitMQMessageConsumer
from .MessageListener import MessageListener as RabbitMQMessageListener
from .MessageParameters import MessageParameters as RabbitMQMessageParameters
from .MessageProducer import MessageProducer as RabbitMQMessageProducer
from .MessageType import MessageType as RabbitMQMessageType
from .Queue import Queue as RabbitMQQueue

__all__ = [
    "RabbitMQBrokerComponent",
    "RabbitMQConnectionSettings",
    "MessageException",
    "RabbitMQExchange",
    "RabbitMQExchangeListener",
    "RabbitMQMessageListener",
    "RabbitMQMessageConsumer",
    "RabbitMQMessageParameters",
    "RabbitMQMessageProducer",
    "RabbitMQMessageType",
    "RabbitMQQueue",
]



from typing import Any, Union
from uuid import UUID as UUIDRaw

from returns.result import Result, Success, safe

from dino_seedwork_be.exceptions import IllegalArgumentException

from .AbstractIdentity import AbstractIdentity

# __all__ = ["UUID"]


class UUID(AbstractIdentity[UUIDRaw]):
    def __init__(self, an_id: Union[UUIDRaw, str]):
        match an_id:
            case UUIDRaw() as id:
                super().__init__(id)
            case str(id):
                super().__init__(UUIDRaw(id))
            case _:
                raise IllegalArgumentException("id should be UUID or UUID string")

    def __eq__(self, obj):
        match obj:
            case UUID():
                return self.id() == obj.id()
            case _:
                return False

    @safe
    def set_id(self, an_id: UUIDRaw | str):
        match an_id:
            case str(id):
                self._id = UUIDRaw(id)
            case UUIDRaw() as id:
                self._id = id

    def validate(self, an_id: UUIDRaw) -> Result[UUIDRaw, Any]:
        return Success(an_id)



import re
from enum import Enum
from typing import Any, List, TypeVar, Union

from returns.future import FutureFailure, FutureResult, FutureSuccess
from returns.iterables import Fold
from returns.maybe import Maybe, Nothing, Some
from returns.pipeline import flow
from returns.pointfree import alt
from returns.result import Result, Success

from dino_seedwork_be.exceptions import (IllegalArgumentException,
                                         IllegalStateException, MainException)
from dino_seedwork_be.utils.functional import return_v


class AssertionErrorCode(Enum):
    ARG_CANNOT_EQUAL = "CANNOT_EQUAL"
    ARG_CANNOT_BE_FALSE = "CANNOT_BE_FALSE"
    ARG_CANNOT_BE_TRUE = "CANNOT_BE_TRUE"
    ARG_LENGTH_INVALID = "LENGTH_INVALID"
    ARG_STRING_CANNOT_EMPTY = "STRING_CANNOT_EMPTY"
    ARG_CANNOT_BE_NONE = "CANNOT_BE_NONE"
    ARG_SHOULD_EQUALS = "NOT_EQUALS"
    ARG_NOT_EQUALS = "NOT_EQUALS"
    ARG_NOT_LARGER_THAN = "NOT_LARGER_THAN"
    ARG_NOT_SMALLER_THAN = "NOT_SMALLER_THAN"
    ARG_NOT_IN_RANGE = "NOT_IN_RANGE"
    ARG_NOT_REGEX_MATCHED = "NOT_REGEX_MATCHED"
    STATE_NOT_SATISFIED = "STATE_NOT_STATISFIED"
    STATE_CANNOT_STATISFIED = "STATE_CANNOT_STATISFIED"


InputType = TypeVar("InputType")


class AssertionConcern:
    IllegalArgExceptionCls = IllegalArgumentException
    IllegalStExceptionCls = IllegalStateException

    @staticmethod
    def simple_handle_on_bool_val(a_bool: bool, exception: Maybe[Exception] = Nothing):
        """
        Throw if a_bool is True
        """
        match a_bool:
            case True:
                return Result.from_failure(exception.value_or(MainException()))
        return Result.from_value("OK")

    def assert_argument_equals(
        self,
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
        an_obj1: object = None,
        an_obj2: object = None,
    ) -> Result:
        is_not_equal = an_obj1 != an_obj2
        return AssertionConcern.simple_handle_on_bool_val(
            is_not_equal,
            exception
            or self.IllegalArgExceptionCls(
                message=a_message.value_or(""),
                code=code.value_or(AssertionErrorCode.ARG_SHOULD_EQUALS.value),
                loc=loc.value_or([]),
            ),
        )

    def assert_argument_false(
        self,
        a_boolean: bool,
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
    ) -> Result:
        is_false = not a_boolean
        return AssertionConcern.simple_handle_on_bool_val(
            not is_false,
            exception
            or self.IllegalArgExceptionCls(
                f"Must be False: {a_message.value_or('')}",
                code=code.value_or(AssertionErrorCode.ARG_CANNOT_BE_TRUE.value),
                loc=loc.value_or([]),
            ),
        )

    def assert_argument_true(
        self,
        a_boolean: bool,
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
    ) -> Result:
        is_true = a_boolean
        return AssertionConcern.simple_handle_on_bool_val(
            not is_true,
            exception
            or self.IllegalArgExceptionCls(
                f"Must be True: {a_message.value_or('')}",
                code=code.value_or(AssertionErrorCode.ARG_CANNOT_BE_FALSE.value),
                loc=loc.value_or([]),
            ),
        )

    def assert_argument_length(
        self,
        a_string: str,
        a_minimum: int,
        a_maximum: int,
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
    ) -> Result:
        length = len(a_string.strip())
        match length > a_maximum:
            case True:
                return Result.from_failure(
                    exception
                    or self.IllegalArgExceptionCls(
                        f"Length must be less than {a_maximum} but {a_string} is not: {a_message}",
                        code=code.value_or(AssertionErrorCode.ARG_LENGTH_INVALID.value),
                        loc=loc.value_or([]),
                    ),
                )

        match length < a_minimum:
            case True:
                return Result.from_failure(
                    exception
                    or self.IllegalArgExceptionCls(
                        f"Length must be large than {a_minimum}: {a_message}",
                        code=code.value_or(AssertionErrorCode.ARG_LENGTH_INVALID.value),
                        loc=loc.value_or([]),
                    )
                )
        return Success(a_string)

    def assert_argument_not_empty(
        self,
        a_string: Maybe[str],
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
    ) -> Result:
        is_empty = a_string.map(
            lambda a_string: a_string is None or len(a_string.strip()) == 0
        ).value_or(False)
        return AssertionConcern.simple_handle_on_bool_val(
            a_bool=is_empty,
            exception=exception
            or self.IllegalArgExceptionCls(
                f"String not empty: {a_message.value_or('')}",
                code=code.value_or(AssertionErrorCode.ARG_STRING_CANNOT_EMPTY.value),
                loc=loc.value_or([]),
            ),
        ).map(return_v(a_string))

    def assert_argument_not_equals(
        self,
        an_object1: object,
        an_object2: object,
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
    ) -> Result:
        return AssertionConcern.simple_handle_on_bool_val(
            an_object1 == an_object2,
            exception
            or self.IllegalArgExceptionCls(
                f"Argument cannot be equal: {a_message}",
                code=code.value_or(AssertionErrorCode.ARG_NOT_EQUALS.value),
                loc=loc.value_or([]),
            ),
        )

    def assert_argument_not_null(
        self,
        an_object: InputType,
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
    ) -> Result[InputType, Any]:
        return AssertionConcern.simple_handle_on_bool_val(
            an_object is None,
            exception
            or self.IllegalArgExceptionCls(
                f"Argument cannot be null: {a_message}",
                code=code.value_or(AssertionErrorCode.ARG_CANNOT_BE_NONE.value),
                loc=loc.value_or([]),
            ),
        ).map(return_v(an_object))

    def assert_argument_larger_than(
        self,
        a_value: Union[int, float],
        a_minium: Union[int, float],
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        allow_equal: bool = False,
        code: Maybe[str] = Nothing,
    ) -> Result:
        return AssertionConcern.simple_handle_on_bool_val(
            (not allow_equal and a_value <= a_minium)
            or (allow_equal and a_value < a_minium),
            exception.lash(
                return_v(
                    Some(
                        self.IllegalArgExceptionCls(
                            f"Argument must be in larger than {a_minium} with {a_value}: {a_message} - loc: {loc}",
                            code=code.value_or(
                                AssertionErrorCode.ARG_NOT_LARGER_THAN.value
                            ),
                            loc=loc.value_or([]),
                        )
                    )
                )
            ),
        ).map(return_v(a_value))

    def assert_argument_smaller_than(
        self,
        a_value: Union[int, float],
        a_maximum: Union[int, float],
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        allow_equal: bool = False,
        code: Maybe[str] = Nothing,
    ) -> Result:
        return AssertionConcern.simple_handle_on_bool_val(
            a_value >= a_maximum or (allow_equal and a_value > a_maximum),
            exception
            or self.IllegalArgExceptionCls(
                f"Argument must be in smaller than {a_maximum}: {a_message} - loc: {loc}",
                code=code.value_or(AssertionErrorCode.ARG_NOT_SMALLER_THAN.value),
                loc=loc.value_or([]),
            ),
        ).map(return_v(a_value))

    def assert_argument_range(
        self,
        a_value: float,
        a_minium: float,
        a_maximum: float,
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
    ) -> Result:
        return AssertionConcern.simple_handle_on_bool_val(
            a_value < a_minium or a_value > a_maximum,
            exception
            or self.IllegalArgExceptionCls(
                f"Value {a_value} just in range {a_minium} -> {a_maximum}: {a_message} - loc: {loc}",
                code=code.value_or(AssertionErrorCode.ARG_NOT_IN_RANGE.value),
                loc=loc.value_or([]),
            ),
        ).map(return_v(a_value))

    def assert_argument_regex(
        self,
        a_value: str,
        regex: str,
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
    ) -> Result:
        return AssertionConcern.simple_handle_on_bool_val(
            re.match(regex, a_value) is None,
            exception.lash(
                return_v(
                    Some(
                        self.IllegalArgExceptionCls(
                            f"Argument does not match the required regex pattern: {a_message} with value {a_value} - loc: {loc}",
                            code=code.value_or(
                                AssertionErrorCode.ARG_NOT_REGEX_MATCHED.value
                            ),
                            loc=loc.value_or([]),
                        )
                    )
                )
            ),
        ).map(return_v(a_value))

    def assert_state_true(
        self,
        a_boolean: bool,
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
    ):
        return AssertionConcern.simple_handle_on_bool_val(
            not a_boolean,
            exception
            or self.IllegalStExceptionCls(
                a_message.value_or(""),
                code=code.value_or(AssertionErrorCode.STATE_NOT_SATISFIED.value),
                loc=loc.value_or([]),
            ),
        )

    def assert_state_false(
        self,
        a_boolean: bool,
        a_message: Maybe[str] = Nothing,
        exception: Maybe[Exception] = Nothing,
        loc: Maybe[List[str]] = Nothing,
        code: Maybe[str] = Nothing,
    ):
        return AssertionConcern.simple_handle_on_bool_val(
            a_boolean,
            exception
            or self.IllegalStExceptionCls(
                a_message.value_or(""),
                code=code.value_or(AssertionErrorCode.STATE_CANNOT_STATISFIED.value),
                loc=loc.value_or([]),
            ),
        )

    def assert_future_maybe_not_nothing(
        self,
        future_maybe: Maybe,
        code: Maybe[str] = Nothing,
        a_message: Maybe[str] = Nothing,
        loc: Maybe[List[str]] = Nothing,
    ) -> FutureResult:
        match future_maybe:
            case Some(sth):
                return FutureSuccess(sth)
            case _:
                return FutureFailure(
                    self.IllegalArgExceptionCls(
                        code=code.value_or(AssertionErrorCode.ARG_CANNOT_BE_NONE.value),
                        message=a_message.value_or(""),
                        loc=loc.value_or([]),
                    )
                )

    def assert_all_not_nothing(
        self,
        a_list: List[Maybe],
        code: Maybe = Nothing,
        a_message: Maybe = Nothing,
        loc: Maybe[List[str]] = Nothing,
    ):
        return flow(
            map(lambda an_item: self.assert_future_maybe_not_nothing(an_item), a_list),
            lambda iter: Fold.collect(iter, Success(())),
            alt(
                return_v(
                    self.IllegalArgExceptionCls(
                        message=a_message.value_or(""),
                        code=code.value_or("ERROR"),
                        loc=loc.value_or([]),
                    )
                )
            ),
        )



import typing as t

T = t.TypeVar("T", bound=t.Callable)


def is_callable_type(tp) -> bool:
    origin = getattr(tp, "__origin__", None)
    return origin is t.Callable.__origin__  # type: ignore


class ReturnType(t.Generic[T]):
    def __class_getitem__(cls, tp):
        if is_callable_type(tp):
            if tp is t.Callable:
                raise TypeError("ReturnType should be parameterized")
            else:
                return_type = tp.__args__[-1]
                return return_type
        else:
            raise TypeError("ReturnType's parameter should be Callable")



from abc import abstractmethod
from typing import Any, Generic, TypeVar

from returns.future import FutureResult

from dino_seedwork_be.application.service import AbstractUOWApplicationService

InputType = TypeVar("InputType")
ReturnType = TypeVar("ReturnType")

# __all__ = ["BaseHandler"]


class BaseHandler(Generic[InputType, ReturnType], AbstractUOWApplicationService):
    @abstractmethod
    def handle(self, command: InputType) -> FutureResult[ReturnType, Any]:
        pass

    def execute(self, command: InputType) -> FutureResult[ReturnType, Any]:
        return self.handle(command)



from abc import abstractmethod
from typing import Any, Generic, Type, TypeVar
from uuid import uuid4

from returns.future import FutureResult, FutureSuccess, future_safe
from returns.maybe import Maybe
from returns.pipeline import pipe
from returns.result import safe
from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio.session import AsyncSession
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.functions import count

from dino_seedwork_be.adapters.persistance.sql.DBSessionUser import \
    DBSessionUser
from dino_seedwork_be.domain.value_object.AbstractIdentity import \
    AbstractIdentity
from dino_seedwork_be.domain.value_object.UUID import UUID
from dino_seedwork_be.repository.IRepository import EntityType, Repository
from dino_seedwork_be.repository.Mapper import Mapper

from .BaseModel import UUIDBaseModel

__all__ = ["StandardAlchemyRepository"]


AlchemyModel = TypeVar("AlchemyModel", bound=UUIDBaseModel)


class StandardAlchemyRepository(
    Generic[EntityType, AlchemyModel],
    Repository[EntityType],
    DBSessionUser[AsyncSession],
):
    model: AlchemyModel
    mapper: Type[Mapper[EntityType, AlchemyModel]]

    def get_next_id(self, _: bool = True) -> FutureResult[AbstractIdentity, Any]:
        return FutureSuccess(UUID(uuid4()))

    def get_by_id(self, id: AbstractIdentity) -> FutureResult[Maybe[EntityType], Any]:
        stmt = select(self.model).where(self.model.id == id.get_raw())
        return (
            future_safe(self.session().execute)(stmt)
            .map(lambda result: result.scalars().first())
            .bind(pipe(self.mapper.to_aggregate, FutureResult.from_result))
        )

    def add(self, entity: EntityType):
        return self.mapper.to_db(entity).map(
            pipe(safe(self.session().add), FutureResult.from_result)
        )

    @future_safe
    async def save(self, entity: EntityType):
        stmt = select(self.model).where(
            self.model.id
            == entity.identity().map(lambda id: id.get_raw()).value_or(None)
        )

        def update(session: Session):
            db_ins = session.execute(stmt).scalars().first()
            self.save_update(db_ins)

        await self.session().run_sync(update)

    @safe
    @abstractmethod
    def save_update(self, db_ins: AlchemyModel):
        pass

    @future_safe
    async def remove(self, id: AbstractIdentity):
        stmt = delete(self.model).where(self.model.id == id.get_raw())
        await self.session().execute(stmt)

    @future_safe
    async def count(self):
        stmt = select(count(self.model.id))

        return (await self.session().execute(stmt)).scalars().first() or 0



import logging
from abc import ABC, abstractmethod
from typing import Callable, Optional

from pika import SelectConnection
from pika.channel import Channel
from pika.connection import ConnectionParameters
from returns.functions import tap
from returns.maybe import Maybe, Nothing, Some
from returns.result import Result

from dino_seedwork_be.exceptions import MainException

from .ConnectionSettings import ConnectionSettings

# I am an abstract class for all Brocker Component, i include an channel
# that all broker component need to refer to

LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)

LOGGER = logging.getLogger(__name__)

# __all__ = ["BrokerComponent"]


class BrokerComponent(ABC):
    # My channel
    _channel: Optional[Channel] = None

    # My connection, which is the connection to my host broker
    _connection: SelectConnection

    # My durable prop, which indicates whether or not messages are durable
    _durable: bool = False

    _connection_settings: ConnectionSettings

    _is_open: bool = False

    _on_setup_finish: Optional[Callable[..., Result]]

    _name: str

    def __init__(
        self,
        a_name: str,
        a_connection_settings: Optional[ConnectionSettings] = None,
        a_broker_channel: Optional["BrokerComponent"] = None,
        on_setup_finish: Optional[Callable[..., Result]] = None,
    ) -> None:
        self.set_name(a_name)
        self._on_setup_finish = on_setup_finish
        match [a_connection_settings, a_broker_channel]:
            case [ConnectionSettings() as connection_setting, None]:
                self.set_connection_settings(connection_setting)
                connection = self._factory_connection(connection_setting)
                self.set_connection(connection)
            case [None, BrokerComponent() as broker_component]:
                """
                Initialize a broker component that reuse connection, channel, and setting
                from an existed component, it is useful for creating a queue that routed by
                an ready exchange
                """
                self.set_connection_settings(broker_component.connection_settings())
                self.set_connection(broker_component.connection())
                broker_component.channel().map(tap(self.set_channel))
                self.setup(self._on_setup_finish)
            case _:
                raise MainException("[Exchange] Params for construtor are not invalid")

    def _factory_connection(
        self,
        a_connection_settings: ConnectionSettings,
    ) -> SelectConnection:
        credentials = a_connection_settings.get_credential()
        con_params = ConnectionParameters(
            host=a_connection_settings.host_name(),
            port=a_connection_settings.port().value_or(
                ConnectionParameters.DEFAULT_PORT
            ),
            virtual_host=a_connection_settings.virtual_host(),
            credentials=credentials.value_or(ConnectionParameters.DEFAULT_CREDENTIALS),
        )

        return SelectConnection(
            con_params,
            on_open_callback=self._on_connection_open,
            on_open_error_callback=self._on_connection_open_error,
            on_close_callback=self._on_connection_closed,
        )

    # TODO: refactor logger info here
    def _on_connection_open(self, _):
        LOGGER.info("Connected success")
        self._open_channel()

    def _on_connection_open_error(self, _, err: Exception):
        LOGGER.error("Connection open failed, reopenning in 5 seconds: %s", err)
        self.connection().ioloop.call_later(5, self.connection().ioloop.stop)

    def _on_connection_closed(self, _, reason):
        self.set_channel(None)
        match self.is_open():
            case True:
                LOGGER.warning(
                    "Connection closed unexpected, reopening in 5s: %s", reason
                )
                self.connection().ioloop.call_later(5, self.connection().ioloop.stop)
            case False:
                LOGGER.info("Deliberately close ioloop on connection closed")
                self.connection().ioloop.stop()

    def _open_channel(self):
        """This method will open a new channel with RabbitMQ by issuing the
        Channel.Open RPC command. When RabbitMQ confirms the channel is open
        by sending the Channel.OpenOK RPC reply, the on_channel_open method
        will be invoked.
        """
        LOGGER.info("Creating a new channel")
        self.connection().channel(on_open_callback=self._on_channel_open)

    def _on_channel_open(self, channel: Channel):
        LOGGER.info("Channel opened")
        self.set_channel(channel)

        LOGGER.info("Adding channel close callback")
        channel.add_on_close_callback(self._on_channel_closed)

        LOGGER.info("Setup detail ...")

        self.setup(self._on_setup_finish)

    def _on_channel_closed(self, channel: Channel, reason: Exception):
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.
        :param pika.channel.Channel channel: The closed channel
        :param Exception reason: why the channel was closed
        """
        LOGGER.warning("Channel %i was closed: %s", channel, reason)
        self.set_channel(None)
        match self.is_open():
            case True:
                self.connection().close()

    @abstractmethod
    def setup(self, callback: Optional[Callable[..., Result]]):
        pass

    def host(self) -> str:
        return self.connection_settings().host_name()

    def port(self) -> Maybe[int]:
        return self.connection_settings().port()

    def connection_settings(self) -> ConnectionSettings:
        return self._connection_settings

    def set_connection_settings(self, connection_settings: ConnectionSettings):
        self._connection_settings = connection_settings

    def name(self) -> str:
        return self._name

    def set_name(self, a_name: str):
        self._name = a_name

    def is_durable(self) -> bool:
        return self._durable

    def set_durable(self, a_bool: bool):
        self._durable = a_bool

    def channel(self) -> Maybe[Channel]:
        return Maybe.from_optional(self._channel)

    def connection(self) -> SelectConnection:
        return self._connection

    def set_channel(self, a_channel: Optional[Channel] = None):
        self._channel = a_channel

    def set_connection(self, a_connection: SelectConnection):
        self._connection = a_connection

    # Answer whether or not I am an exchange channel
    def is_exchange(self) -> bool:
        return False

    # Answers my name as the exchange name if I am an Exchange
    # otherwhise the empty string

    def exchange_name(self) -> Maybe[str]:
        return Some(self.name()) if self.is_exchange() else Nothing

    # Answer whether or not I am an queue channel
    def is_queue(self) -> bool:
        return False

    # Answers my name as the queue name if I am an Queue
    # otherwhise the empty string

    def queue_name(self) -> Maybe[str]:
        return Some(self.name()) if self.is_queue() else Nothing

    def is_open(self) -> bool:
        return self._is_open

    def _set_is_open(self, is_open: bool):
        self._is_open = is_open

    def _close_channel(self):
        """Invoke this command to close the channel with RabbitMQ by sending
        the Channel.Close RPC command.
        """

        def exe_with_unnone(_):
            LOGGER.info("Close the channel")
            self.channel().unwrap().close()

        self.channel().map(tap(exe_with_unnone))

    def _close_connection(self):
        """This method closes the connection to RabbitMQ."""

        self.connection().close()

    def close(self):
        # RabbitMQ doesn't guarantee that if isOpen()
        # answers true that close() will work because
        # another client may be racing to close the
        # same process and/or components. so here just
        # attempt to close, catch and ignore, and move
        # on to next steps is the recommended approach.
        #
        # for the purpose here, the isOpen() checks prevent
        # closing a shared channel and connection that is
        # shared by a subscriber exchange and queue.
        self._set_is_open(False)
        self._close_channel()
        self._close_connection()

    def run(self):
        LOGGER.info("io loop run")
        """Run the example code by connecting and then starting the IOLoop."""
        try:
            self.connection().ioloop.start()
        except Exception:
            # Gracefully close the connection
            self.connection().close()
            # Loop until we're fully closed, will stop on its own
            self.connection().ioloop.start()



import re
from typing import Optional

from returns.maybe import Maybe, Nothing, Some
from returns.result import safe

from dino_seedwork_be.domain.value_object.AbstractValueObject import \
    ValueObject

# __all__ = ["StringWithRegex"]


class StringWithRegex(ValueObject):
    _pattern: str
    _value: Maybe[str] = Nothing
    _error_message: str = "Value does not path the pattern: "

    def __init__(self, a_regex: str, an_error_message: Optional[str] = None):
        self.set_pattern(a_regex)
        if an_error_message is not None:
            self._error_message = an_error_message

    @safe
    def set_value(
        self,
        a_value: Maybe[str],
        exception_code: Maybe[str] = Nothing,
        loc: Maybe[list[str]] = Nothing,
    ):
        match a_value:
            case Some(str(v)):
                self.assert_state_true(
                    re.match(self._pattern, v) is not None,
                    Some(f"{self._error_message} {a_value}"),
                    code=exception_code,
                    loc=loc,
                ).unwrap()
                self._value = a_value
            case Maybe.empty:
                self._value = Nothing

    def set_pattern(self, a_regex: str):
        self.assert_argument_not_empty(
            Some(a_regex), Some("Pattern for a StringRegex cannot be empty")
        ).unwrap()
        self._pattern = a_regex

    def value(self) -> Maybe[str]:
        return self._value



from abc import ABC, abstractmethod


class JSONSerializable(ABC):
    @abstractmethod
    def as_dict(self) -> dict:
        pass

    def __getstate__(self):
        return self.as_dict()

    @staticmethod
    @abstractmethod
    def restore(a_dict: dict):
        ...

    def get_value(self):
        return self.__getstate__()
